﻿using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;

namespace Sudoku
{
    public partial class frmSudoku : Form
    {
        private struct sCell
        {
            public int Value;
            public int Row;
            public int Col;
            public int Sec;
            public bool Fixed;
        }
        private struct sNum
        {
            public int Value;
            public int Col;
            public int ColMask;
            public int SecMask;
        }
        private sCell[,] _sudoku = null;
        private int[] _numbers = null;
        private sNum[,,] _matrix = null;

        public frmSudoku()
        {
            InitializeComponent();
        }

        public int GetNumber(int value)
        {
            switch (value)
            {
                case 0:
                    return 0;
                case 1:
                    return 1;
                case 2:
                    return 2;
                case 4:
                    return 3;
                case 8:
                    return 4;
                case 16:
                    return 5;
                case 32:
                    return 6;
                case 64:
                    return 7;
                case 128:
                    return 8;
                default:
                    return 9;
            }
        }
        private Bitmap DrawSudoku()
        {
            Bitmap bmp = new Bitmap(640, 480);
            Graphics gr = Graphics.FromImage(bmp);
            gr.FillRectangle(Brushes.White, new Rectangle(0, 0, 640, 480));
            for (int ix = 1; ix < 9; ix++)
            {
                gr.DrawLine(Pens.Gray, 95 + (ix * 50), 15, 95 + (ix * 50), 465);
                gr.DrawLine(Pens.Gray, 95, 15 + (ix * 50), 545, 15 + (ix * 50));
            }
            Pen p = new Pen(Brushes.Black, 3f);
            gr.DrawRectangle(p, new Rectangle(95, 15, 450, 450));
            gr.DrawLine(p, 245, 15, 245, 465);
            gr.DrawLine(p, 395, 15, 395, 465);
            gr.DrawLine(p, 95, 165, 545, 165);
            gr.DrawLine(p, 95, 315, 545, 315);
            p.Dispose();

            Font f = new Font("Arial", 40f, FontStyle.Bold, GraphicsUnit.Pixel);
            for (int row = 0; row < 9; row++)
            {
                for (int col = 0; col < 9; col++)
                {
                    if (GetNumber(_sudoku[row, col].Value) > 0)
                    {
                        if (_sudoku[row, col].Fixed)
                        {
                            gr.DrawString(GetNumber(_sudoku[row, col].Value).ToString(), f, Brushes.Red, 100 + (col * 50), 20 + (row * 50));
                        }
                        else
                        {
                            gr.DrawString(GetNumber(_sudoku[row, col].Value).ToString(), f, Brushes.Black, 100 + (col * 50), 20 + (row * 50));
                        }
                    }
                }
            }
            f.Dispose();
            gr.Dispose();
            return bmp;
        }
        private void SetCell(int row, int col, int v, bool fix)
        {
            _sudoku[row, col].Value = v;
            _sudoku[row, col].Fixed = fix;
            for (int c = 0; c < 9; c++)
            {
                _sudoku[row, c].Row |= v;
                _sudoku[c, col].Col |= v;
                _sudoku[(3 * (row / 3)) + (c % 3), (3 * (col / 3)) + (c / 3)].Sec |= v;
            }
        }
        private void BuildMatrix()
        {
            int v = 1;
            for (int n = 0; n < 9; n++)
            {                
                for (int row = 0; row < 9; row++)
                {
                    int c = 0;
                    int msec = 0x3F;
                    int mcol = 0x7EFF;
                    int bit = 0x100;
                    for (int col = 0; col < 9; col++)
                    {
                        if (_sudoku[row, col].Fixed)
                        {
                            if (_sudoku[row, col].Value == v)
                            {
                                _matrix[row, c, n].Value = bit;
                                _matrix[row, c, n].Col = col;
                                _matrix[row, c, n].ColMask = mcol & 0x1FF;
                                _matrix[row, c++, n].SecMask = msec;
                            }
                        }
                        else if (((_sudoku[row, col].Row | 
                            _sudoku[row, col].Col | 
                            _sudoku[row, col].Sec) & v) == 0)
                        {
                            _matrix[row, c, n].Value = bit;
                            _matrix[row, c, n].Col = col;
                            _matrix[row, c, n].ColMask = mcol & 0x1FF;
                            _matrix[row, c++, n].SecMask = msec;
                        }
                        bit >>= 1;
                        mcol >>= 1;
                        mcol |= 0x4000;
                        if (col == 2)
                        {
                            msec = 0x1C7;
                        }
                        else if (col == 5)
                        {
                            msec = 0x1F8;
                        }
                    }
                }
                v <<= 1;
            }
        }
        private bool ProcessNextNumber(int n)
        {
            int[] indexes = new int[9];
            int[] oldnumbers = new int[9];
            for (int num = 0; num < 9; num++)
            {
                oldnumbers[num] = _numbers[num];
            }
            int mcol = 0x1FF;
            int msec = 0x1FF;
            int ix = 0;
            int v = 1 << n;
            while (true)
            {
                int index = indexes[ix];
                if ((index < 9) &&
                    ((_matrix[ix, index, n].Value & msec & mcol & ~_numbers[ix]) != 0))
                {                    
                    _numbers[ix] |= _matrix[ix, index, n].Value;
                    _sudoku[ix, _matrix[ix, index, n].Col].Value = v;
                    mcol &= _matrix[ix, index, n].ColMask;
                    if ((ix == 2) || (ix == 5))
                    {
                        msec = 0x1FF;
                    }
                    else
                    {
                        msec &= _matrix[ix, index, n].SecMask;
                    }
                    ix++;
                    if (ix == 9)
                    {
                        if (n == 8)
                        {
                            return true;
                        }
                        if (ProcessNextNumber(n + 1))
                        {
                            return true;
                        }
                        ix = 8;                        
                        _numbers[8] = oldnumbers[8];
                        mcol |= (~_matrix[8, indexes[8], n].ColMask) & 0x1FF;
                        msec = _matrix[7, indexes[7], n].SecMask & 
                            _matrix[6, indexes[6], n].SecMask;
                        indexes[8]++;
                    }
                }
                else
                {
                    if ((index < 9) && (_matrix[ix, index, n].Value != 0))
                    {
                        indexes[ix]++;
                    }
                    else
                    {
                        if (ix == 0)
                        {
                            return false;
                        }
                        indexes[ix] = 0;
                        ix--;
                        _numbers[ix] = oldnumbers[ix];
                        mcol |= (~_matrix[ix, indexes[ix], n].ColMask) & 0x1FF;                        
                        switch (ix % 3)
                        {
                            case 0:
                                msec = 0x1FF;
                                break;
                            case 1:
                                msec = _matrix[ix - 1, indexes[ix - 1], n].SecMask;
                                break;
                            default:
                                msec = _matrix[ix - 1, indexes[ix - 1], n].SecMask & 
                                    _matrix[ix - 2, indexes[ix - 2], n].SecMask;
                                break;
                        }
                        indexes[ix]++;
                    }
                }
            }            
        }
        private void bNew_Click(object sender, EventArgs e)
        {
            if (ofDlg.ShowDialog() == DialogResult.OK)
            {
                _sudoku = new sCell[9, 9];
                _matrix = new sNum[9, 9, 9];
                _numbers = new int[9];
                StreamReader rdr = new StreamReader(ofDlg.FileName);
                try
                {
                    for (int row = 0; row < 9; row++)
                    {
                        string[] line = rdr.ReadLine().Split(',');
                        for (int col = 0; col < 9; col++)
                        {
                            int v = int.Parse(line[col]);
                            if (v != 0)
                            {
                                v = 1 << (v - 1);
                                if (((_sudoku[row, col].Row | _sudoku[row, col].Col | _sudoku[row, col].Sec) & v) != 0)
                                {
                                    throw new Exception(string.Format("Bad Sudoku! ({0},{1})", row + 1, col + 1));
                                }
                                SetCell(row, col, v, true);
                            }
                        }
                    }
                    BuildMatrix();
                    if (ProcessNextNumber(0))
                    {
                        pbSudoku.Image = DrawSudoku();
                    }
                    else
                    {
                        throw new Exception("Cannot be solved!");
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                    pbSudoku.Image = DrawSudoku();
                }
                finally
                {
                    rdr.Close();
                }
            }
        }
    }
}
